@@ -163,13 +163,16 @@ typedef struct _AFS_NONPAGED_FCB
     USHORT          Type;
 
     //
-    // Ranking - File Resource first, then Paging Resource
+    // Ranking - File Resource, Paging Resource,
+    //           then SectionObject Resource
     //
 
     ERESOURCE       Resource;
 
     ERESOURCE       PagingResource;
 
+    ERESOURCE       SectionObjectResource;
+
     //
     // The section object pointer
     //
@@ -558,6 +561,32 @@ typedef struct _AFS_DEVICE_EXTENSION
 
             KEVENT              MemoryAvailableEvent;
 
+            //
+            // Worker Pool Queues
+            //
+
+            ERESOURCE           QueueLock;
+
+            struct _AFS_WORK_ITEM     *QueueHead;
+
+            struct _AFS_WORK_ITEM     *QueueTail;
+
+            KEVENT              WorkerQueueHasItems;
+
+            LONG                QueueItemCount;
+
+            ERESOURCE           IOQueueLock;
+
+            struct _AFS_WORK_ITEM     *IOQueueHead;
+
+            struct _AFS_WORK_ITEM     *IOQueueTail;
+
+            KEVENT              IOWorkerQueueHasItems;
+
+            LONG                IOQueueItemCount;
+
+
+
         } Control;
 
         struct
@@ -664,16 +693,6 @@ typedef struct _AFS_DEVICE_EXTENSION
 
             struct _AFS_WORKER_QUEUE_HDR *PoolHead;
 
-            ERESOURCE        QueueLock;
-
-            struct _AFS_WORK_ITEM     *QueueHead;
-
-            struct _AFS_WORK_ITEM     *QueueTail;
-
-            KEVENT           WorkerQueueHasItems;
-
-            LONG             QueueItemCount;
-
             //
             // IO Worker queue
             //
@@ -682,16 +701,6 @@ typedef struct _AFS_DEVICE_EXTENSION
 
             struct _AFS_WORKER_QUEUE_HDR *IOPoolHead;
 
-            ERESOURCE        IOQueueLock;
-
-            struct _AFS_WORK_ITEM     *IOQueueHead;
-
-            struct _AFS_WORK_ITEM     *IOQueueTail;
-
-            KEVENT           IOWorkerQueueHasItems;
-
-            LONG             IOQueueItemCount;
-
         } Library;
 
     } Specific;
@@ -743,6 +752,8 @@ NTSTATUS
                        IN GUID *AuthGroup,
                        IN PUNICODE_STRING FileName,
                        IN AFSFileID *FileId,
+                       IN WCHAR * Cell,
+                       IN ULONG   CellLength,
                        IN void  *Data,
                        IN ULONG DataLength,
                        IN OUT void *ResultBuffer,
