@@ -42,6 +42,7 @@ NTSTATUS
 AFSLoadLibrary( IN ULONG Flags,
                 IN UNICODE_STRING *ServicePath)
 {
+    UNREFERENCED_PARAMETER(Flags);
 
     NTSTATUS ntStatus = STATUS_SUCCESS;
     AFSDeviceExt       *pDevExt = (AFSDeviceExt *)AFSDeviceObject->DeviceExtension;
@@ -477,6 +478,11 @@ AFSClearLibraryRequest()
 
     __Enter
     {
+        AFSDbgLogMsg( AFS_SUBSYSTEM_LOAD_LIBRARY,
+                      AFS_TRACE_LEVEL_VERBOSE,
+                      "%s Inflight Count %08lX\n",
+                      __FUNCTION__,
+                      pDevExt->Specific.Control.InflightLibraryRequests);
 
         if( InterlockedDecrement( &pDevExt->Specific.Control.InflightLibraryRequests) == 0)
         {
@@ -486,11 +492,6 @@ AFSClearLibraryRequest()
                         FALSE);
         }
 
-        AFSDbgLogMsg( AFS_SUBSYSTEM_LOAD_LIBRARY,
-                      AFS_TRACE_LEVEL_VERBOSE,
-                      "%s Inflight Count %08lX\n",
-                      __FUNCTION__,
-                      pDevExt->Specific.Control.InflightLibraryRequests);
     }
 
     return ntStatus;
@@ -848,6 +849,7 @@ NTSTATUS
 AFSInitializeLibrary( IN AFSFileID *GlobalRootFid,
                       IN BOOLEAN QueueRootEnumeration)
 {
+    UNREFERENCED_PARAMETER(QueueRootEnumeration);
 
     NTSTATUS ntStatus = STATUS_SUCCESS;
     AFSLibraryInitCB    stInitLib;
@@ -870,6 +872,8 @@ AFSInitializeLibrary( IN AFSFileID *GlobalRootFid,
 
         stInitLib.AFSServerName = AFSServerName;
 
+        stInitLib.AFSMountRootName = AFSMountRootName;
+
         stInitLib.AFSDebugFlags = AFSDebugFlags;
 
         if( GlobalRootFid != NULL)
@@ -895,7 +899,7 @@ AFSInitializeLibrary( IN AFSFileID *GlobalRootFid,
 
         stInitLib.AFSExAllocatePoolWithTag = AFSExAllocatePoolWithTag;
 
-        stInitLib.AFSExFreePool = AFSExFreePool;
+        stInitLib.AFSExFreePoolWithTag = AFSExFreePoolWithTag;
 
         stInitLib.AFSDumpTraceFiles = AFSDumpTraceFiles;
 
