@@ -1,7 +1,7 @@
 /*
  * Copyright 2000, International Business Machines Corporation and others.
  * All Rights Reserved.
- * 
+ *
  * This software has been released under the terms of the IBM Public
  * License.  For details, see the LICENSE file in the top-level source
  * directory or online at http://www.openafs.org/dl/license10.html
@@ -11,9 +11,14 @@
  * INCLUDES ___________________________________________________________________
  *
  */
+#include <winsock2.h>
+#include <ws2tcpip.h>
+
 extern "C" {
+#include <afsconfig.h>
 #include <afs/param.h>
 #include <afs/stds.h>
+#include <roken.h>
 }
 
 #include "afscfg.h"
@@ -41,9 +46,9 @@ cfg_partitionEntry_t *GetPartitionTable(int &nNumPartitions)
 }
 
 int GetNumPartitions()
-{	
+{
     return cPartitions;
-}	
+}
 
 int ReadPartitionTable(afs_status_t *pStatus)
 {
@@ -53,7 +58,7 @@ int ReadPartitionTable(afs_status_t *pStatus)
 
     cPartitions = 0;
     pTable = 0;
-	
+
     int nResult = cfg_HostPartitionTableEnumerate(g_hServer, &pTable, &cPartitions, pStatus);
 
     return nResult;
@@ -73,7 +78,7 @@ BOOL IsAnAfsPartition(LPCTSTR pszRootDir)
 
 	if (ch1 == ch2)
 	    return TRUE;
-    }	
+    }
 
     return FALSE;
 }
@@ -83,11 +88,11 @@ BOOL DoesPartitionExist(LPCTSTR pszName)
     for (int ii = 0; ii < cPartitions; ii++) {
 	if (lstrcmp(A2S(pTable[ii].partitionName), pszName) == 0)
 	    return TRUE;
-    }	
-	
+    }
+
     return FALSE;
 }
-	
+
 void FreePartitionTable()
 {
     if (pTable) {
@@ -97,5 +102,5 @@ void FreePartitionTable()
 
     pTable = 0;
     cPartitions = 0;
-}	
+}
 
