@@ -42,8 +42,8 @@ afs_create(OSI_VC_DECL(adp), char *aname, struct vattr *attrs,
 {
     afs_int32 origCBs, origZaps, finalZaps;
     struct vrequest treq;
-    register afs_int32 code;
-    register struct afs_conn *tc;
+    afs_int32 code;
+    struct afs_conn *tc;
     struct VenusFid newFid;
     struct AFSStoreStatus InStatus;
     struct AFSFetchStatus OutFidStatus, OutDirStatus;
@@ -56,6 +56,7 @@ afs_create(OSI_VC_DECL(adp), char *aname, struct vattr *attrs,
     struct vcache *tvc;
     struct volume *volp = 0;
     struct afs_fakestat_state fakestate;
+    struct rx_connection *rxconn;
     XSTATS_DECLS;
     OSI_VC_CONVERT(adp);
 
@@ -297,14 +298,14 @@ afs_create(OSI_VC_DECL(adp), char *aname, struct vattr *attrs,
 
     	InStatus.UnixModeBits = attrs->va_mode & 0xffff;	/* only care about protection bits */
     	do {
-	    tc = afs_Conn(&adp->f.fid, &treq, SHARED_LOCK);
+	    tc = afs_Conn(&adp->f.fid, &treq, SHARED_LOCK, &rxconn);
 	    if (tc) {
-	    	hostp = tc->srvr->server;	/* remember for callback processing */
+	    	hostp = tc->parent->srvr->server; /* remember for callback processing */
 	    	now = osi_Time();
 	    	XSTATS_START_TIME(AFS_STATS_FS_RPCIDX_CREATEFILE);
 	    	RX_AFS_GUNLOCK();
 	    	code =
-		    RXAFS_CreateFile(tc->id, (struct AFSFid *)&adp->f.fid.Fid,
+		    RXAFS_CreateFile(rxconn, (struct AFSFid *)&adp->f.fid.Fid,
 				 aname, &InStatus, (struct AFSFid *)
 				 &newFid.Fid, &OutFidStatus, &OutDirStatus,
 				 &CallBack, &tsync);
@@ -314,7 +315,7 @@ afs_create(OSI_VC_DECL(adp), char *aname, struct vattr *attrs,
 	    } else
 	    	code = -1;
     	} while (afs_Analyze
-	         (tc, code, &adp->f.fid, &treq, AFS_STATS_FS_RPCIDX_CREATEFILE,
+	         (tc, rxconn, code, &adp->f.fid, &treq, AFS_STATS_FS_RPCIDX_CREATEFILE,
 	          SHARED_LOCK, NULL));
 
 	if ((code == EEXIST || code == UAEEXIST) &&
@@ -362,12 +363,10 @@ afs_create(OSI_VC_DECL(adp), char *aname, struct vattr *attrs,
 	}
 
     } else {
-#if defined(AFS_DISCON_ENV)
 	/* Generate a fake FID for disconnected mode. */
 	newFid.Cell = adp->f.fid.Cell;
 	newFid.Fid.Volume = adp->f.fid.Fid.Volume;
 	afs_GenFakeFid(&newFid, VREG, 1);
-#endif
     }				/* if (!AFS_IS_DISCON_RW) */
 
     /* otherwise, we should see if we can make the change to the dir locally */
@@ -452,10 +451,8 @@ afs_create(OSI_VC_DECL(adp), char *aname, struct vattr *attrs,
 	    }
 	    ReleaseWriteLock(&afs_xcbhash);
 	    if (AFS_IS_DISCON_RW) {
-#if defined(AFS_DISCON_ENV)
 		afs_DisconAddDirty(tvc, VDisconCreate, 0);
 		afs_GenDisconStatus(adp, tvc, &newFid, attrs, &treq, VREG);
-#endif
 	    } else {
 		afs_ProcessFS(tvc, &OutFidStatus, &treq);
 	    }
@@ -492,6 +489,7 @@ afs_create(OSI_VC_DECL(adp), char *aname, struct vattr *attrs,
 	afs_AddMarinerName(aname, *avcp);
 	/* return the new status in vattr */
 	afs_CopyOutAttrs(*avcp, attrs);
+	afs_MarinerLog("store$Creating", *avcp);
     }
 
     afs_PutFakeStat(&fakestate);
@@ -513,10 +511,10 @@ afs_create(OSI_VC_DECL(adp), char *aname, struct vattr *attrs,
  * and dcache entry write-locked.
  */
 int
-afs_LocalHero(register struct vcache *avc, register struct dcache *adc,
-	      register AFSFetchStatus * astat, register int aincr)
+afs_LocalHero(struct vcache *avc, struct dcache *adc,
+	      AFSFetchStatus * astat, int aincr)
 {
-    register afs_int32 ok;
+    afs_int32 ok;
     afs_hyper_t avers;
 
     AFS_STATCNT(afs_LocalHero);
